from typing import NamedTuple

from parse_qwantz.pixels import Pixel


class Box(NamedTuple):
    top_left: Pixel
    bottom_right: Pixel

    @property
    def top_right(self) -> Pixel:
        return Pixel(self.bottom_right.x, self.top_left.y)

    @property
    def bottom_left(self) -> Pixel:
        return Pixel(self.top_left.x, self.bottom_right.y)

    @property
    def left(self) -> int:
        return self.top_left.x

    @property
    def right(self) -> int:
        return self.bottom_right.x

    @property
    def top(self) -> int:
        return self.top_left.y

    @property
    def bottom(self) -> int:
        return self.bottom_right.y

    @property
    def width(self) -> int:
        return self.right - self.left

    @property
    def height(self) -> int:
        return self.bottom - self.top

    def includes(self, pixel: Pixel) -> bool:
        return self.left <= pixel.x < self.right and self.top <= pixel.y < self.bottom

    def contains(self, box: "Box") -> bool:
        return self.includes(box.top_left) and self.includes(box.bottom_right)

    def distance_squared(self, box: "Box") -> int:
        return (
            get_interval_distance((self.left, self.right), (box.left, box.right))**2
            + get_interval_distance((self.top, self.bottom), (box.top, box.bottom))**2
        )

    @classmethod
    def dummy(cls) -> "Box":
        return cls(Pixel(0, 0), Pixel(0, 0))


def get_interval_distance(interval1: tuple[int, int], interval2: tuple[int, int]) -> int:
    i1, i2 = sorted((interval1, interval2))
    return max(0, i2[0] - i1[1])



import logging
import re
from itertools import chain, groupby
from typing import Iterable, NamedTuple

from parse_qwantz.box import Box, get_interval_distance
from parse_qwantz.fonts import Font, CharBox
from parse_qwantz.colors import Color
from parse_qwantz.text_lines import TextLine, try_text_line
from parse_qwantz.pixels import Pixel
from parse_qwantz.simple_image import SimpleImage

logger = logging.getLogger()


class TextBlock(NamedTuple):
    lines: list[TextLine]
    bond_strengths: list[int]
    color: Color
    font: Font

    @property
    def start(self) -> Pixel:
        return self.lines[0].start

    @property
    def end(self) -> Pixel:
        return self.lines[-1].end

    @property
    def is_bold(self) -> bool:
        return all(line.is_bold for line in self.lines)

    @property
    def box(self) -> Box:
        top = self.start.y
        bottom = self.end.y
        left = min(line.start.x for line in self.lines)
        right = max(line.end.x for line in self.lines)
        return Box(Pixel(left, top), Pixel(right, bottom))

    def content(self, mark_bold=True, include_font_name=False):
        char_boxes = []
        for line in self.lines:
            if char_boxes:
                if char_boxes[-1].char != '-' or char_boxes[-2].char == ' ':
                    char_boxes.append(CharBox.space(is_bold=char_boxes[-1].is_bold))
                else:
                    logger.warning(f"Line ending with '-', ambiguous ({line.content.split()[-1]})")
            char_boxes.extend(line.char_boxes)

        grouped_char_boxes = groupby(char_boxes, key=lambda cb: cb.is_bold and mark_bold)
        text_and_weight = (
            (''.join(char_box.char for char_box in group), is_bold)
            for is_bold, group in grouped_char_boxes
        )
        content = ''.join(
            make_bold_excluding_trailing_spaces(content) if is_bold else content
            for content, is_bold in text_and_weight
        )
        content = content.replace('  ', ' ')
        if self.font.italic_offsets:
            content = f'_{content}_'
        if include_font_name and self.font.name not in ('Regular', 'Italic'):
            content = f"({self.font.name.lower()}) {content}"
        return content

    def split(self, line1: TextLine, line2: TextLine) -> tuple["TextBlock", "TextBlock"]:
        line1_index = self.lines.index(line1)
        line2_index = self.lines.index(line2)
        index1, index2 = sorted((line1_index, line2_index))
        _, split_index = min((self.bond_strengths[i], i) for i in range(index1, index2))
        block1 = TextBlock(self.lines[:split_index+1], self.bond_strengths[:split_index], self.color, self.font)
        block2 = TextBlock(self.lines[split_index+1:], self.bond_strengths[split_index+1:], self.color, self.font)
        if line1_index < line2_index:
            return block1, block2
        else:
            return block2, block1

    def __str__(self):
        return self.content()

    def __hash__(self):
        return id(self)


def make_bold_excluding_trailing_spaces(s: str) -> str:
    trailing_spaces = re.search(r" *$", s).group()
    return f'**{s.rstrip()}**{trailing_spaces}'


def get_text_blocks(text_lines: list[TextLine], image: SimpleImage) -> Iterable[TextBlock]:
    grouped_lines = group_text_lines(text_lines)
    text_lines = sorted(
        chain.from_iterable(join_text_lines(group, image) for group in grouped_lines),
        key=lambda l: (l.start.y, l.start.x)
    )
    while text_lines:
        new_block = [text_lines[0]]
        bond_strengths = []
        font = text_lines[0].font
        new_lines: list[TextLine] = []
        for text_line in text_lines[1:]:
            append = False
            bond_strength = 0
            if text_line.font == font:
                text_box = text_line.box()
                previous_line = new_block[-1]
                bold_mismatch = previous_line.is_bold and not text_line.contains_bold
                if not bold_mismatch:
                    previous_box = previous_line.box()
                    intervals_intersect = get_interval_distance(
                        (text_box.left, text_box.right),
                        (previous_box.left, previous_box.right),
                    ) == 0
                    ceiling = previous_box.bottom
                    previous_height = previous_line.font.height
                    previous_width = previous_line.font.width
                    if ceiling - 1 <= text_box.top <= ceiling + previous_height // 6 and intervals_intersect:
                        if previous_box.left == text_box.left:
                            bond_strength += 5
                        elif (previous_box.left - text_box.left) % previous_width == 0:
                            bond_strength += 3
                        if text_box.top <= ceiling:
                            bond_strength += 10
                        if text_box.top >= ceiling + previous_height:
                            bond_strength -= 10
                        append = True
            if append:
                new_block.append(text_line)
                bond_strengths.append(bond_strength)
            else:
                new_lines.append(text_line)
        text_lines = new_lines
        found_pixel = new_block[0].find_pixel(image)
        color = image.pixels[found_pixel] if found_pixel else Color.WHITE
        yield TextBlock(new_block, bond_strengths, color, font)


def group_text_lines(text_lines: list[TextLine]) -> list[list[TextLine]]:
    grouped_text_lines = []
    used: set[TextLine] = set()
    for text_line in text_lines:
        if text_line in used:
            continue
        used.add(text_line)
        group = [text_line]
        box = text_line.box()
        for other_text_line in text_lines:
            if other_text_line in used:
                continue
            other_box = other_text_line.box()
            if abs(box.top - other_box.top) <= 1 or abs(box.bottom - other_box.bottom) <= 1:
                distance = other_box.left - box.right
                if -1 <= distance <= max(group[-1].font.width, other_text_line.font.width) * 2 + 1:
                    group.append(other_text_line)
                    used.add(other_text_line)
        grouped_text_lines.append(group)
    return grouped_text_lines


def join_text_lines(text_lines: list[TextLine], image: SimpleImage) -> list[TextLine]:
    if len(text_lines) == 1:
        return text_lines
    for font in (line.font for line in text_lines):
        if font == text_lines[0].font:
            continue
        first_pixel = text_lines[0].find_pixel(image)
        if joined_text_line := try_text_line(first_pixel, image, font):
            joined_box = joined_text_line.box()
            if abs(joined_box.right - text_lines[-1].box().right) < font.width // 2:
                return [joined_text_line]
    return text_lines



from logging import getLogger

from parse_qwantz.box import Box
from parse_qwantz.fonts import ALL_FONTS
from parse_qwantz.lines import Line, get_line
from parse_qwantz.text_lines import TextLine, try_text_line
from parse_qwantz.detect_thought import get_thought
from parse_qwantz.pixels import Pixel
from parse_qwantz.shape import get_shape
from parse_qwantz.simple_image import SimpleImage

logger = getLogger()


def get_elements(image: SimpleImage) -> tuple[list[Line], list[Box], list[TextLine], list[list[Pixel]]]:
    text_lines: list[TextLine] = []
    lines: list[Line] = []
    thoughts: list[Box] = []
    unmatched: list[list[Pixel]] = []
    sorted_pixels = sorted(image.pixels)
    while sorted_pixels:
        pixel = sorted_pixels[0]
        pixels = set(sorted_pixels)
        tmp_image = SimpleImage(
            width=image.width,
            height=image.height,
            pixels={pixel: color for pixel, color in image.pixels.items() if pixel in pixels},
        )
        for font in ALL_FONTS:
            text_line = try_text_line(pixel, tmp_image, font)
            if text_line:
                text_lines.append(text_line)
                if text_line.font.italic_offsets:
                    sorted_pixels = remove_italic(sorted_pixels, text_line)
                else:
                    sorted_pixels = remove_boxes(sorted_pixels, [char_box.box for char_box in text_line.char_boxes])
                break
        else:
            result = get_line(pixel, tmp_image)
            if result:
                line, line_pixels = result
                lines.append(line)
                sorted_pixels = remove_subsequence(sorted_pixels, line_pixels)
            elif result := get_thought(pixel, tmp_image):
                box, thought_pixels = result
                thoughts.append(box)
                sorted_pixels = remove_subsequence(sorted_pixels, thought_pixels)
            else:
                logger.warning(f"No match found for shape at {(pixel.x, pixel.y)}")
                unmatched_pixels = sorted(get_shape(pixel, tmp_image))
                unmatched.append(unmatched_pixels)
                sorted_pixels = remove_subsequence(sorted_pixels, unmatched_pixels)
    return lines, thoughts, sorted(text_lines, key=lambda l: l.start), unmatched


def remove_boxes(sorted_pixels: list[Pixel], boxes: list[Box]) -> list[Pixel]:
    box_iter = iter(boxes)
    box = next(box_iter)
    new_pixels = []
    pass_through = False
    for pixel in sorted_pixels:
        while not pass_through and pixel.x >= box.right:
            try:
                box = next(box_iter)
            except StopIteration:
                pass_through = True
        if pass_through or not box.includes(pixel):
            new_pixels.append(pixel)
    return new_pixels


def remove_italic(sorted_pixels: list[Pixel], text_line: TextLine) -> list[Pixel]:
    italic_pixels = []
    for char_box in text_line.char_boxes:
        italic_pixels.extend(
            char_box.pixels(italic_offsets=text_line.font.italic_offsets)
        )
    return sorted(set(sorted_pixels) - set(italic_pixels))


def remove_subsequence(sorted_pixels: list[Pixel], subsequence: list[Pixel]) -> list[Pixel]:
    sub_iter = iter(subsequence)
    next_pixel = next(sub_iter)
    pixels = []
    for pixel in sorted_pixels:
        while next_pixel is not None and pixel > next_pixel:
            try:
                next_pixel = next(sub_iter)
            except StopIteration:
                next_pixel = None
        if pixel != next_pixel:
            pixels.append(pixel)
    return pixels



from logging import getLogger
from pathlib import Path
from typing import Iterable

from PIL import Image, ImageDraw

from parse_qwantz.box import Box, get_interval_distance
from parse_qwantz.color_logs import set_logging_formatter
from parse_qwantz.colors import Color
from parse_qwantz.text_blocks import get_text_blocks, TextBlock
from parse_qwantz.elements import get_elements
from parse_qwantz.match_blocks import match_blocks, Character_s
from parse_qwantz.match_lines import match_lines, Character, OFF_PANEL, UnmatchedLine
from parse_qwantz.match_thought import match_thought
from parse_qwantz.pixels import Pixel
from parse_qwantz.shape import get_box
from parse_qwantz.simple_image import SimpleImage
from parse_qwantz.prepare_image import prepare_image
from parse_qwantz.text_lines import TextLine

set_logging_formatter()

logger = getLogger()

# size and offset
PANELS = [
    ((239, 239), (3, 2)),
    ((126, 239), (246, 2)),
    ((358, 239), (375, 2)),
    ((190, 239), (3, 244)),
    ((295, 239), (196, 244)),
    ((239, 239), (494, 244)),
]

CHARACTERS = [
    [Character('T-Rex', Box(Pixel(104, 90), Pixel(170, 238)))],
    [Character('T-Rex', Box(Pixel(4, 119), Pixel(105, 238)))],
    [
        Character('T-Rex', Box(Pixel(80, 55), Pixel(115, 213))),
        Character('Dromiceiomimus', Box(Pixel(325, 150), Pixel(357, 238))),
        Character('House', Box(Pixel(115, 210), Pixel(163, 238)), can_think=False),
    ],
    [
        Character('T-Rex', Box(Pixel(0, 65), Pixel(30, 190))),
        Character('Utahraptor', Box(Pixel(103, 81), Pixel(138, 165))),
        Character('Girl', Box(Pixel(0, 213), Pixel(8, 238)), can_think=False),
    ],
    [
        Character('T-Rex', Box(Pixel(40, 70), Pixel(96, 103))),
        Character('Utahraptor', Box(Pixel(200, 80), Pixel(233, 145))),
    ],
    [Character('T-Rex', Box(Pixel(74, 64), Pixel(120, 195)))],
]


def parse_qwantz(image: Image, debug: bool) -> Iterable[list[str]]:
    masked = prepare_image(image)
    for (panel, characters) in zip(PANELS, CHARACTERS):
        (width, height), (x, y) = panel
        cropped = masked.crop((x, y, x + width, y + height))
        panel_image = SimpleImage.from_image(cropped)
        try:
            script_lines, unmatched_shapes, unmatched_neighbors = parse_panel(panel_image, characters)
            if debug:
                draw = ImageDraw.Draw(cropped)
                for unmatched_shape in unmatched_shapes:
                    box = get_box(unmatched_shape, padding=3)
                    draw.rectangle(box, outline=(255, 0, 0))
                    for pixel in unmatched_shape:
                        draw.point(pixel, fill=(255, 0, 0))
                for text_line1, text_line2 in unmatched_neighbors:
                    box1 = text_line1.box()
                    box2 = text_line2.box()
                    draw.rectangle(box1, outline=(0, 0, 192))
                    draw.rectangle(box2, outline=(0, 0, 192))
                    draw.line([
                        ((box1.left + box1.right) // 2, (box1.top + box1.bottom) // 2),
                        ((box2.left + box2.right) // 2, (box2.top + box2.bottom) // 2),
                    ], fill=(0, 0, 255))
                if unmatched_shapes or unmatched_neighbors:
                    cropped.show()
            yield script_lines
        except UnmatchedLine as e:
            line, boxes, text_blocks = e.args
            if debug:
                draw = ImageDraw.Draw(cropped)
                for box, _character in boxes:
                    draw.rectangle(box, outline=(0, 192, 0))
                draw.line(line, fill=(255, 0, 0))
                cropped.show()
            yield ["Error"] + text_blocks


def parse_panel(
    image: Image, characters: list[Character]
) -> tuple[list[str], list[list[Pixel]], list[tuple[TextLine, TextLine]]]:
    lines, thoughts, text_lines, unmatched = get_elements(image)
    text_blocks = sorted(get_text_blocks(text_lines, image), key=lambda b: (b.end.y, b.end.x))
    line_matches = match_lines(lines, text_blocks, characters, image)
    block_matches, text_blocks, unmatched_neighbors = match_blocks(line_matches, text_blocks)
    text_blocks = sorted(text_blocks, key=lambda b: (b.end.y, b.end.x))
    unmatched_blocks = [block for block in text_blocks if block not in block_matches]
    thinking_characters = [character for character in characters if character.can_think]
    thought_matches = {
        block: character
        for block, character in match_thought(thoughts, unmatched_blocks, thinking_characters)
    }
    if thoughts and not thought_matches:
        logger.warning("Detected thought bubbles, but no thought text")
    unmatched_blocks = [block for block in unmatched_blocks if block not in thought_matches]
    match_above_or_below(unmatched_blocks, block_matches)
    script_lines = []
    for block in text_blocks:
        if god_or_devil := handle_god_and_devil(block, block_matches.get(block) == OFF_PANEL):
            block_matches[block] = god_or_devil
        if block in block_matches:
            character = block_matches[block]
            if isinstance(character, tuple):
                script_lines.append(f"{character[0]} and {character[1]}: {block.content(include_font_name=True)}")
            elif character.name in ('God', 'Devil'):
                script_lines.append(f"{character}: {block.content(mark_bold=False)}")
            else:
                script_lines.append(f"{character}: {block.content(include_font_name=True)}")
        elif block in thought_matches:
            character = thought_matches[block]
            script_lines.append(f"{character}: (thinks) {block.content()}")
        else:
            if not block.is_bold:
                logger.warning('Narrator not bold: %s', block.font.name)
            script_lines.append(f"Narrator: {block.content(mark_bold=False)}")
    return script_lines, unmatched, unmatched_neighbors


def match_above_or_below(unmatched_blocks: list[TextBlock], block_matches: dict[TextBlock, Character_s]) -> None:
    for unmatched_block in unmatched_blocks:
        box = unmatched_block.box
        closest = None
        best_distance = None
        if not unmatched_block.is_bold:
            for block, character in block_matches.items():
                other_box = block.box
                if get_interval_distance((box.left, box.right), (other_box.left, other_box.right)) == 0:
                    distance = max(other_box.top - box.bottom, box.top - other_box.bottom)
                    line_height = max(unmatched_block.font.height, block.font.height)
                    if distance < line_height * 2 and (best_distance is None or distance < best_distance):
                        closest = character
                        best_distance = distance
        if closest:
            block_matches[unmatched_block] = closest


def handle_god_and_devil(block: TextBlock, is_off_panel: bool):
    if block.color == Color.RED:
        if not is_off_panel:
            logger.warning('Red block not off-panel')
        if block.font.name != 'Bold':
            logger.warning('Red block not bold')
        return Character.from_name('Devil')
    elif is_off_panel and block.font.name == 'Bold':
        return Character.from_name('God')


def main(input_file_path: str | Path, debug: bool):
    for panel_no, panel in enumerate(parse_qwantz(Image.open(input_file_path), debug=debug), start=1):
        print(f'Panel {panel_no}:')
        for line in panel:
            print(line)
        if panel_no != 6:
            print()



import logging
from dataclasses import dataclass
from typing import Iterable

from parse_qwantz.box import Box
from parse_qwantz.lines import Line
from parse_qwantz.text_blocks import TextBlock
from parse_qwantz.pixels import Pixel
from parse_qwantz.simple_image import SimpleImage
from parse_qwantz.text_lines import TextLine

logger = logging.getLogger()


@dataclass
class Character:
    name: str
    box: Box
    can_think: bool = True

    def __str__(self):
        return self.name

    @classmethod
    def from_name(cls, name: str):
        return cls(name, Box.dummy())


OFF_PANEL = Character.from_name("Off-Panel")

Target = TextLine | Character


class UnmatchedLine(Exception):
    pass


def match_lines(
    lines: list[Line], text_blocks: list[TextBlock], characters: list[Character], image: SimpleImage
) -> Iterable[tuple[Target, Target]]:
    boxes: list[tuple[Box, Target]] = [
        (text_line.box(padding=1), text_line)
        for text_block in text_blocks
        for text_line in text_block.lines
    ]
    boxes.extend((character.box, character) for character in characters)
    for line in lines:
        end1, end2 = line
        distance1 = distance2 = None
        closest1 = closest2 = None
        if image.is_on_edge(end1):
            distance1 = 0
            closest1 = OFF_PANEL
        if image.is_on_edge(end2):
            distance2 = 0
            closest2 = OFF_PANEL
        for box, target in boxes:
            for side in sides(box):
                if intersects(line, side):
                    t = relative_distance_to_intersection(line, side)
                    if t <= 0 and (distance1 is None or -t < distance1):
                        distance1 = -t
                        closest1 = target
                    elif t > 0 and (distance2 is None or t < distance2):
                        distance2 = t
                        closest2 = target
                    break
        if not (isinstance(closest1, TextLine) or isinstance(closest2, TextLine)) or closest1 is closest2 is None:
            logger.error(f"Unmatched line {line}: matches {closest1} to {closest2}")
            raise UnmatchedLine(line, boxes, text_blocks)
        if (closest1 is None and closest2 is not None) or (closest1 is not None and closest2 is None):
            logger.warning(f"Unmatched line {line}, assuming off-panel")
            if closest1 is None:
                closest1 = OFF_PANEL
            if closest2 is None:
                closest2 = OFF_PANEL
        if closest1 == closest2:
            logger.error(f"Line {line} matches the same object: {closest1}")
            raise UnmatchedLine(line, boxes, text_blocks)
        yield closest1, closest2


def sides(box: Box) -> list[Line]:
    return [
        (box.top_left, box.bottom_left),
        (box.bottom_left, box.bottom_right),
        (box.bottom_right, box.top_right),
        (box.top_right, box.top_left),
    ]


def intersects(line: Line, segment: Line) -> bool:
    (x0, y0), (x1, y1) = line
    (ax, ay), (bx, by) = segment
    return ((y0 - y1)*(ax - x0) + (x1 - x0)*(ay - y0)) * ((y0 - y1)*(bx - x0) + (x1 - x0)*(by - y0)) < 0


def relative_distance_to_intersection(line: Line, segment: Line) -> float:
    (x0, y0), (x1, y1) = line
    (ax, ay), (bx, by) = segment
    # (x0, y0) + t (x1 - x0, y1 - y0)
    # -> for which t is it on the ab line?
    t = ((ax - x0)*(by - ay) - (ay - y0)*(bx - ax)) / ((x1 - x0)*(by - ay) - (y1 - y0)*(bx - ax))
    if t > 1 or t < 0:
        return t
    else:
        return -t if is_left(line, segment[0]) else 2 - t


def is_left(line: Line, point: Pixel) -> bool:
    (ax, ay), (bx, by) = line
    cx, cy = point
    return (bx - ax)*(cy - ay) - (by - ay)*(cx - ax) > 0



import logging
from typing import Iterable

from parse_qwantz.text_blocks import TextBlock
from parse_qwantz.match_lines import Target, Character
from parse_qwantz.text_lines import TextLine

logger = logging.getLogger()

Character_s = Character | tuple[Character, Character]


def match_blocks(
    line_matches: Iterable[tuple[Target, Target]], text_blocks: list[TextBlock]
) -> tuple[dict[TextBlock, Character_s], list[TextBlock], list[tuple[TextLine, TextLine]]]:
    block_matches: dict[TextBlock, tuple[Character_s, TextLine]] = {}
    blocks_by_line: dict[TextLine, TextBlock] = {line: block for block in text_blocks for line in block.lines}
    neighbors: list[tuple[TextLine, TextLine]] = []
    for target1, target2 in line_matches:
        if isinstance(target1, TextLine) and isinstance(target2, TextLine):
            neighbors.append((target1, target2))
        else:
            character, line = (target1, target2) if isinstance(target1, Character) else (target2, target1)
            block = blocks_by_line[line]
            if block not in block_matches:
                block_matches[block] = (character, line)
            else:
                prev_character, prev_line = block_matches[block]
                if prev_line == line:
                    block_matches[block] = ((prev_character, character), line)
                else:
                    block1, block2 = block.split(prev_line, line)
                    for b1line in block1.lines:
                        blocks_by_line[b1line] = block1
                    for b2line in block2.lines:
                        blocks_by_line[b2line] = block2
                    del block_matches[block]
                    block_matches[block1] = (prev_character, prev_line)
                    block_matches[block2] = (character, line)
    while neighbors:
        neighbors_left = []
        for line1, line2 in neighbors:
            if blocks_by_line[line1] == blocks_by_line[line2]:
                block = blocks_by_line[line1]
                block1, block2 = block.split(line1, line2)
                for b1line in block1.lines:
                    blocks_by_line[b1line] = block1
                for b2line in block2.lines:
                    blocks_by_line[b2line] = block2
                if block in block_matches:
                    character, line = block_matches[block]
                    del block_matches[block]
                    block_matches[block1] = (character, line)
                    block_matches[block2] = (character, line)
            else:
                block1 = blocks_by_line[line1]
                block2 = blocks_by_line[line2]
            if block1 in block_matches:
                block_matches[block2] = block_matches[block1]
            elif block2 in block_matches:
                block_matches[block1] = block_matches[block2]
            else:
                neighbors_left.append((line1, line2))
        if neighbors == neighbors_left:
            logger.warning(f"Unmatched connected text lines: {neighbors}")
            break
        neighbors = neighbors_left
    final_block_matches = {block_id: character for block_id, (character, line) in block_matches.items()}
    text_blocks = list(set(blocks_by_line.values()))
    return final_block_matches, text_blocks, neighbors



from pathlib import Path

import typer

from parse_qwantz.main import main

app = typer.Typer()


@app.command()
def cli(image_path: Path, debug: bool = typer.Option(False, help="Enable debug features.")):
    """Generate transcripts for Ryan North's Dinosaur Comics from https://qwantz.com"""
    main(image_path, debug=debug)


if __name__ == '__main__':
    app()



import string
from dataclasses import dataclass
from importlib.resources import as_file, files
from itertools import product
from pathlib import Path
from typing import NamedTuple, ContextManager

from PIL import Image

import parse_qwantz
from parse_qwantz.box import Box
from parse_qwantz.pixels import Pixel
from parse_qwantz.simple_image import SimpleImage

PRINTABLE = string.printable.strip()
FORBIDDEN_CHARS = '\\_`|~'

FONT_SIZES = [(13, 'Regular'), (12, 'Condensed'), (11, 'Small'), (9, 'Mini'), (8, 'Tiny')]
SHIFTED_VARIANTS = {
    13: {',': 1, ':': 1, '.': -1},
}


class CharBox(NamedTuple):
    char: str
    box: Box
    is_bold: bool

    def pixels(self, italic_offsets: set[int]):
        if not italic_offsets:
            return [
                (x, y)
                for x in range(self.box.left, self.box.right)
                for y in range(self.box.top, self.box.bottom)
            ]
        else:
            pixels = []
            italic_offset = len(italic_offsets)
            for y in range(self.box.top, self.box.bottom):
                if y - self.box.top in italic_offsets:
                    italic_offset -= 1
                for x in range(self.box.left, self.box.right):
                    pixels.append((x + italic_offset, y))
            return pixels

    @classmethod
    def space(cls, is_bold: bool) -> "CharBox":
        return cls(char=' ', box=Box.dummy(), is_bold=is_bold)


@dataclass
class Font:
    name: str
    width: int
    height: int
    shapes: dict[int, str]
    bold_shapes: dict[int, str]
    italic_offsets: set[int]
    top_pixel: int
    left_pixel: int

    def get_char(
        self,
        pixel: Pixel,
        image: SimpleImage,
        expect_bold: bool = False,
        expect_space: bool = True,
    ) -> CharBox | None:
        if char_box := self.get_char_with_weight(pixel, image, is_bold=expect_bold, expect_space=expect_space):
            return char_box
        return self.get_char_with_weight(pixel, image, is_bold=not expect_bold, expect_space=expect_space)

    def get_char_with_weight(
        self,
        pixel: Pixel,
        image: SimpleImage,
        is_bold: bool,
        expect_space: bool,
    ) -> CharBox | None:
        width = self.width + 1 if is_bold else self.width
        bottom_right = Pixel(pixel.x + width, pixel.y + self.height)
        bitmask = self._get_bitmask(pixel, image, is_bold)
        if char := self._get_char_by_bitmask(bitmask, is_bold):
            return CharBox(char, Box(pixel, bottom_right), is_bold)

        for cut_bottom in range(1, 3):
            cut_bitmask = bitmask & -(1 << (width * cut_bottom))
            if cut_bitmask == 0 or cut_bitmask & -cut_bitmask > (1 << (width * (cut_bottom + 1))):
                if char := self._get_char_by_bitmask(cut_bitmask, is_bold):
                    right, bottom = bottom_right
                    return CharBox(char, Box(pixel, Pixel(right, bottom - cut_bottom)), is_bold)

        for cut_top in range(1, 2):
            cut_bitmask = bitmask & ((1 << (width * (self.height - cut_top))) - 1)
            if cut_bitmask == 0 or cut_bitmask < (1 << (width * (self.height - cut_top - 1))):
                if char := self._get_char_by_bitmask(cut_bitmask, is_bold):
                    return CharBox(char, Box(Pixel(pixel.x, pixel.y + cut_top), bottom_right), is_bold)

        if expect_space:
            right, bottom = bottom_right
            for x, y in product(range(pixel.x, right), range(pixel.y, bottom)):
                if Pixel(x, y) in image.pixels:
                    if x >= right - width // 2 - 1:
                        return CharBox(' ', Box(pixel, Pixel(x, bottom)), is_bold)
                    break

    def _get_bitmask(self, pixel: Pixel, image: SimpleImage, is_bold: bool) -> int:
        width = self.width + 1 if is_bold else self.width
        return get_bitmask(pixel, image, width, self.height, self.italic_offsets)

    def _get_char_by_bitmask(self, bitmask: int, is_bold: bool) -> str | None:
        if bitmask == 0:
            return ' '
        shapes = self.bold_shapes if is_bold else self.shapes
        char = shapes.get(bitmask)
        if char and char not in FORBIDDEN_CHARS:
            return char

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Font(name={self.name}, width={self.width}, height={self.height})"

    @classmethod
    def from_file(
        cls,
        file_path_context_manager: ContextManager[Path],
        name: str,
        italic_offsets: set[int],
        shifted_variants: dict[str, int] | None = None,
    ) -> "Font":
        with file_path_context_manager as file_path:
            image = SimpleImage.from_image(Image.open(file_path))
        width = image.width // len(PRINTABLE)
        height = image.height
        shapes = {}
        for i, char in enumerate(PRINTABLE):
            bitmask = get_bitmask(
                Pixel(width * i, 0), image=image, width=width, height=height, italic_offsets=italic_offsets
            )
            shapes[bitmask] = char
            if shifted_variants and char in shifted_variants:
                shapes[get_shifted_variant(bitmask, width, height, shifted_variants[char])] = char
            cut_bitmask = bitmask & -(1 << width)
            if cut_bitmask != bitmask and char not in 'gq[]':
                shapes[cut_bitmask] = char
        top_pixel = max(get_top_pixel(bitmask, width, height) for bitmask in shapes)
        left_pixel = max(get_left_pixel(bitmask, width, height) for bitmask in shapes)
        return cls(
            name,
            width,
            height,
            shapes,
            get_bold_shapes(width, height, shapes),
            italic_offsets,
            top_pixel,
            left_pixel,
        )


def get_top_pixel(bitmask: int, width: int, height: int) -> int:
    for j in range(height):
        if bitmask & (((1 << width) - 1) << (width * (height - j - 1))):
            return j
    return 0


def get_left_pixel(bitmask: int, width: int, height: int) -> int:
    mask = sum(1 << (width * j) for j in range(height))
    for j in range(width):
        if bitmask & (mask << (width - j - 1)):
            return j
    return 0


def get_shifted_variant(shape: int, width: int, height: int, offset: int) -> int:
    shifted = 0
    mask = (1 << width) - 1
    for level in range(height):
        line = shape & mask
        if offset >= 0:
            shifted_line = line >> offset
        else:
            shifted_line = line << -offset
        shifted |= shifted_line << (level * width)
        shape >>= width
    return shifted


def get_bold_shapes(width: int, height: int, shapes: dict[int, str]) -> dict[int, str]:
    return {
        regular_shape_to_bold(shape, width, height): char
        for shape, char in shapes.items()
    }


def regular_shape_to_bold(shape: int, width: int, height: int) -> int:
    bold = 0
    mask = (1 << width) - 1
    for level in range(height):
        line = shape & mask
        bold_line = line | (line << 1)
        bold |= bold_line << (level * (width + 1))
        shape >>= width
    return bold


def get_bitmask(
    pixel: Pixel, image: SimpleImage, width: int, height: int, italic_offsets: set[int]
) -> int:
    bitmask = 0
    x0, y0 = pixel
    italic_offset = len(italic_offsets)
    for y in range(y0, y0 + height):
        if y - y0 in italic_offsets:
            italic_offset -= 1
        for x in range(x0, x0 + width):
            bitmask <<= 1
            if (x + italic_offset, y) in image.pixels:
                bitmask += 1
    return bitmask


ALL_FONTS = [
    Font.from_file(
        file_path_context_manager=as_file(files(parse_qwantz).joinpath(f'img/regular{size}.png')),
        name=name,
        italic_offsets=set(),
        shifted_variants=SHIFTED_VARIANTS.get(size, {}),
    )
    for size, name in FONT_SIZES
]

ALL_FONTS.append(
    Font.from_file(
        file_path_context_manager=as_file(files(parse_qwantz).joinpath(f'img/italic13.png')),
        name='Italic',
        italic_offsets={3, 5, 9, 11},
        shifted_variants={},
    )
)



from collections import deque
from itertools import product

from parse_qwantz.box import Box
from parse_qwantz.pixels import Pixel
from parse_qwantz.simple_image import SimpleImage
from parse_qwantz.shape import get_shape, get_adjacent_pixels, get_box


def get_thought(pixel: Pixel, image: SimpleImage) -> tuple[Box, list[Pixel]] | None:
    orig_pixels = get_shape(pixel, image)
    box = get_box(orig_pixels)
    pixels: set[Pixel] = set(orig_pixels)
    for x, y in orig_pixels:
        pixels.update({Pixel(x + 1, y), Pixel(x + 2, y)})
    x_range = range(box.left + 1, box.right - 1)
    y_range = range(box.top + 1, box.bottom - 1)
    outside = set()
    outside.update(Pixel(box.left, y) for y in y_range)
    outside.update(Pixel(box.right, y) for y in y_range)
    outside.update((Pixel(x, box.top - 1) for x in x_range))
    outside.update((Pixel(x, box.bottom - 1) for x in x_range))
    non_empty_interior = False
    for x, y in product(x_range, y_range):
        if Pixel(x, y) in pixels:
            continue
        visited = set()
        to_visit = deque([Pixel(x, y)])
        while to_visit:
            current = to_visit.popleft()
            if current in visited or current in pixels or not box.includes(current):
                continue
            if current in outside:
                outside.update(visited)
                break
            visited.add(current)
            to_visit.extend(get_adjacent_pixels(current))
        else:
            non_empty_interior = True
            break
    if non_empty_interior:
        return box, sorted(orig_pixels)
